import hashlib
import sys
import os
from config import SEC_DB_PATH, DISTRO
import json
import re

# VALID_PARENTS = [
#     "/usr/local/lib/security/pam_opendirectory_typo.so",
#     "/sbin/unix_chkpwd
# ]

def get_ppid_and_attr(pid):
    cmd = "ps -p %d -oppid=,uid=,user=,comm=" % pid
    try:
        ppid, uid, user, exe = re.split(r'\s+', os.popen(cmd).read().strip(), maxsplit=3)
    except Exception as e:
        print e
        ppid, uid, user, exe = '-1', '', '', ''
    return ppid, uid, user, exe

def load_recoreded_digest():
    return [
        h.split()[0] for h in
        open(os.path.join(SEC_DB_PATH, 'authorized_caller'), 'r')\
        .read().strip().split('\n')
        if h
    ]

def sha256(fname):
    if not os.path.exists(fname):
        return "-1"
    hash_sha256 = hashlib.sha256()
    with open(fname, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_sha256.update(bytes(chunk))
    return hash_sha256.hexdigest()


def is_valid_parent():
    """
    Authenticates the script by validating top 3 parents, if any of
    them belongs to VALID_PARENTS, with matching RECORDED_DIGEST.
    """
    # f = open('/tmp/typtop.log', 'a')
    RECORDED_DIGESTS = load_recoreded_digest()
    ppid = os.getppid()
    for _ in xrange(3):
        ppid, uid, user, exe = get_ppid_and_attr(ppid)
        if not ppid or int(ppid) <= 0: break
        ppid = int(ppid)
        continue;
        if uid and int(uid) == 0: # any of the uids is 0 (root)
            return True
        if sha256(exe) in RECORDED_DIGESTS:
            return True
    # f.close()
    return False

def validate_pam_opendirectory(fname):
    return sha256(fname) in RECOREDED_DIGESTS

if __name__ == '__main__':
    print load_recoreded_digest()
    print sha256(sys.argv[1])
    print is_valid_parent()



from __future__ import print_function
import os, sys
import argparse
from typtop.dbaccess import (
    UserTypoDB,
    call_check
)
from typtop.config import (
    SEC_DB_PATH, NUMBER_OF_ENTRIES_TO_ALLOW_TYPO_LOGIN,
    WARM_UP_CACHE, VERSION, GROUP, DISTRO, BINDIR, first_msg,
)
from typtop.validate_parent import is_valid_parent
import subprocess

USER = ""
SEND_LOGS_SCRIPT = '{}/send_typo_log.py'.format(BINDIR)
if not os.path.exists(SEND_LOGS_SCRIPT):
    SEND_LOGS_SCRIPT = '/usr/bin/send_typo_log.py'
    if not os.path.exists(SEND_LOGS_SCRIPT):
        SEND_LOGS_SCRIPT = 'send_typo_log.py'

ALLOW_TYPO_LOGIN = True


class AbortSettings(RuntimeError):
    pass

def _get_login_user():
    # gets the username of the logging user
    pp = subprocess.Popen('who', stdout=subprocess.PIPE)
    output = pp.stdout.read()
    first_line = output.splitlines()[0]
    user = first_line.split()[0]
    return user


def _get_username():
    # trying to go over the problem of
    if USER:
        print("Designated user: {}".format(USER))
        return USER
    uid = os.getuid()
    is_root = uid == 0
    user = _get_login_user()
    if is_root:
        r = raw_input("Setting will be done for login user: {}.\n"
                      "Please confirm. (Yn) ".format(user))
        abort = r and r.lower() == "n"
        if abort:
            raise AbortSettings()
    else:
        print("Designated user: {}".format(user))
    return user


def _get_typoDB():
    user = _get_username()
    try:
        typoDB = UserTypoDB(user)
    except Exception as e:
        print(
            "It seems you have not initialized the db. Try running"\
            " \"sudo {} --init\" to initialize the db.\nThe error "\
            "I ran into is the following:\n{}"
            .format(sys.argv[0], e)
        )
        return None
    if not typoDB.is_typtop_init():
        raise Exception("{}:{} not initiated".format(
            str(typoDB),
            typoDB.get_db_path())
        )
    return typoDB


def root_only_operation():
    if os.getuid() != 0:
        print("ERROR!! You need root privilege to run this operation")
        raise AbortSettings


def initiate_typodb(RE_INIT=False):
    # ValueError(
    #     "You should not require to call this. "
    #     "Something is wrong!! Try re-installing the whole system"
    # )
    root_only_operation()
    # user = _get_username()
    # try:
    #     # checks that such a user exists:
    #     _ = pwd.getpwnam(user).pw_dir
    # except KeyError as e:
    #     print("Error: {}".format(e.message))
    #     print("Hint: The user ({}) must have an account in this computer."\
    #           .format(user))
    #     print("Hint 2: It's not a registration. User the username for "\
    #           "your account in the computer.")
    if False:
        pass
    else:
        branch = "master"
        subdir, download_bin, makecmd = '', '', ''
        if DISTRO == 'darwin':
            # TODO: Cleanup this directories. e.g., pam_opendirectory
            subdir = 'csrcs/osx/prebuilt'
            download_bin = "curl -LO"
            makecmd = './install.sh'
        elif DISTRO in ('debian', 'fedora', 'arch'):
            subdir = 'csrcs/linux/'
            download_bin = "wget"
            makecmd = "make && make install"
        cmd = """
        cd /tmp/ && {download_bin} https://github.com/rchatterjee/pam-typopw/archive/{branch}.zip && unzip -qq {branch}.zip \
        && cd pam-typopw-{branch}/{subdir} && {makecmd};
        cd /tmp && rm -rf {branch}.zip pam-typopw*
        """.format(branch=branch, subdir=subdir,
                   download_bin=download_bin, makecmd=makecmd)
        print(cmd)
        os.system(cmd)

common_auth = {   # Not used
    'debian': '/etc/pam.d/common-auth',
    'fedora': '/etc/pam.d/system-auth',
    'darwin': '',
    'arch'  : '/etc/pam.d/system-auth',
}[DISTRO]


def uninstall_pam_typtop():
    # Last try to send logs
    root_only_operation()
    typtop_uninstall_script = BINDIR + '/typtop-uninstall.sh'
    print(DISTRO)
    subprocess.call(typtop_uninstall_script)

parser = argparse.ArgumentParser("typtop ")
# parser.add_argument(
#     "--user",
#     help="To set the username. Otherwise login user will be the target"
# )

parser.add_argument(
    "--init", action="store_true",
    help="To initialize the DB. You have to run this once you install pam_typtop"
)

parser.add_argument(
    "--allowtypo", type=str.lower, choices=['yes','no'],
    help='Allow login with typos of the password'
)

parser.add_argument(
    "--allowupload", type=str.lower, choices=['yes', 'no'],
    help="Allow uploading the non-sensitive anonymous "\
    "data into the server for research purposes."
)

# parser.add_argument(
#     "--installid", action="store_true",
#     help="Prints the installation id, which you have to submit while filling up the google form"
# )

parser.add_argument(
    "--status", action="store", nargs="+",
    help='Prints current states of the typo-tolerance. Needs a username as argument.'
)

parser.add_argument(
    "--uninstall", action="store_true",
    help="Uninstall TypToP from your machine. Will delete all the data related to TypTop too."
)

# parser.add_argument(
#     "--reinit", action="store_true",
#     help="To re-initiate the DB, especially after the user's pw has changed"
# )

parser.add_argument(
    "--update", action="store_true",
    help="Updates TypTop to the latest released version"
)

parser.add_argument(
    "--check", action="store", nargs=3,
    help="(INTERNAL FUNCTION. PLEASE DON'T CALL THIS.)"
)

parser.add_argument(
    "--debug", action="store_true",
    help="Prepare report for debugging"
)

def main():
    args = parser.parse_args()
    if len(sys.argv) <=1:
        print(parser.print_help())
        sys.exit(0)

    # ITS IMPORTANT THIS ONE WILL BE FIRST
    # if args.user:
    #     global USER
    #     USER = args.user
        # print("User settings have been set to {}".format(USER))
    try:
        # root_only_operation()
        if args.allowtypo:
            typoDB = _get_typoDB()
            if args.allowtypo == "no":
                typoDB.allow_login(False)
                print(
                    """
Turning OFF login with typos. The software will still monitor your
typos and build cache of popular typos. You can switch on this
whenever you want.
                    """)  # :{}".format(typoDB.is_allowed_login())
            elif args.allowtypo == "yes":
                print("Turning ON login with typos...",)
                typoDB.allow_login(True)

        if args.allowupload:
            typoDB = _get_typoDB()
            if args.allowupload == "yes":
                typoDB.allow_upload(True)
                print("Uploading data is enabled. You are awesome. Thanks!!")
            elif args.allowupload == "no":
                typoDB.allow_upload(False)
                print("Uploading data is disabled.  :( :'( :-(!")
                print("Thanks for using the software anyway.")

        if args.init:
            print(first_msg, file=sys.stderr)
            print("Initializing the typo database..")
            initiate_typodb()

        # if args.reinit:
        #     print("RE-initiating pam_typtop")
        #     initiate_typodb(RE_INIT=True)

        if args.status:
            users = args.status
            if not users:
                users.add(_get_username())
            for user in users:
                typoDB = UserTypoDB(user)
                print("\n** TYPO-TOLERANCE STATUS **\n")
                print(">> User: {}".format(user))
                print("\tLogin with typos: {}".format(typoDB.is_allowed_login()))
                print("\tParticipate in the study: {}"\
                      .format(typoDB.is_allowed_upload()))
                print("\tIs enough logins to allow typos: {}"\
                      .format(typoDB.check_login_count(update=False)))
                print("\tInstall Id: {}".format(typoDB.get_installation_id().strip()))
                print("\tSoftware Version: {}".format(VERSION))
                print("\tNum entries before typo-login allowed: {}".format(NUMBER_OF_ENTRIES_TO_ALLOW_TYPO_LOGIN))
                print("\tWarmup cache: {}".format(WARM_UP_CACHE))

        if args.uninstall:
            r = raw_input("Uninstalling pam_typtop. Will delete all the "
                          "databases.\nPlease confirm. (yN)")
            if r and r.lower() == "y":
                uninstall_pam_typtop()

        if args.update:  # delete all old data
            cmd =  """export PIP_FORMAT=columns;
                pip list --outdated|grep typtop;
                if [ "$?" = "0" ]; then
                   pip uninstall typtop
                   pip install -U --ignore-installed typtop && typtops.py --init
                else
                   echo "Already uptodate! No need to update."
                fi
            """
            os.system(cmd)

        if args.debug:
            p = subprocess.Popen(
                'pip show numpy', shell=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT
            )
            numpypath = ''
            for l in p.stdout:
                l = l.strip().split(': ', 1)
                if l[0] == 'Location':
                    numpypath = l[1]
                    break
            print("Is numpy in path: {}".format(numpypath in sys.path))
            proc = subprocess.Popen(
                # TODO: Add numpy path constraint

                """
                set -x
                set -u

                users
                # su $USER -c "which su"
                # <enter correct password>
                # if [ $? -neq 0 ]; then exit; else "echo password incorrect"; fi
                typtop --status $USER
                sudo ls -altrh  /usr/local/etc/typtop.d/$USER/typtop.json
                ls -altrh $(which su) $(which typtop)
                python -c "import pwd; print pwd.getpwnam('$USER')"
                tail -n50 /var/log/typtop.log
                """, shell=True,
                stdout=subprocess.PIPE,
                # stderr=subprocess.STDOUT
                stderr=sys.stdout.fileno()
            )
            print(proc.stdout.read())

        if args.check:
            # ensure the parent is pam_opendirectory_typo.so
            assert is_valid_parent()
            failed, user, pw = args.check
            ret = call_check(failed, user, pw)
            sys.stdout.write(str(ret))
            # if ret==0:
            #     p = subprocess.Popen([SEND_LOGS_SCRIPT, user])

    except AbortSettings:
        print("Settings' change had been aborted.")


if __name__ == '__main__':
    main()



#!/usr/bin/env python2.7

import os
import time
import json
import yaml
import sys
import pwd
import random
from zxcvbn import password_strength
from collections import defaultdict
from base64 import urlsafe_b64encode, urlsafe_b64decode
from word2keypress import distance
from operator import itemgetter

# Local libraries
from typtop.dbutils import (
    logger, setup_logger, is_user, get_machine_id, is_in_top5_fixes,
    change_db_ownership
)
from typtop.config import (
    DB_NAME, auxT, INSTALLATION_ID,
    INSTALLATION_DATE, LOG_LAST_SENTTIME, LOG_SENT_PERIOD,
    UPDATE_GAPS, SYSTEM_STATUS, SYSTEM_STATUS_NOT_INITIALIZED,
    LOGIN_COUNT, ALLOWED_TYPO_LOGIN, ALLOWED_UPLOAD, ENC_PK, INDEX_J,
    WAITLIST_SIZE, WARM_UP_CACHE, CACHE_SIZE, REAL_PW, HMAC_SALT,
    FREQ_COUNTS, HEADER_CTX, SYSTEM_STATUS_ALL_GOOD, LOWER_ENT_CUTOFF,
    EDIT_DIST_CUTOFF, WAIT_LIST, TYPO_CACHE, SEC_DB_PATH,
    NUMBER_OF_ENTRIES_TO_ALLOW_TYPO_LOGIN, REL_ENT_CUTOFF,
    SYSTEM_STATUS_PW_CHANGED, SYSTEM_STATUS_CORRUPTED_DB, LOG_DIR,
    logT, GROUP, warm_up_with
)
from typtop.pw_pkcrypto import (
    generate_key_pair, compute_id,
    pkencrypt, pkdecrypt, pwencrypt, pwdecrypt,
    serialize_pk, deserialize_pk, serialize_sk, deserialize_sk,
    verify_pk_sk, SALT_LENGTH
)

# GENERAL TODO:
# - improve computation speed, moved to json file, but still slow
#

_entropy_cache = {}


def get_logging_path(username):
    homedir = pwd.getpwnam(username).pw_dir
    return "{}/{}.log".format(homedir, DB_NAME)


def get_time():
    """
    Returns the timestamp in a string, in a consistent format
    which works in linux and can be stored in the DB
    (unlike datetime.datetime, for example)
    """
    return time.time()


def entropy(typo):
    if typo not in _entropy_cache:
        _entropy_cache[typo] = password_strength(typo)['entropy']
    return _entropy_cache[typo]


class UserTypoDB(object):
    class TypoDBError(Exception):
        # all errors that have to do with the typoDB state
        pass

    class NoneInitiatedDB(TypoDBError):
        pass

    class CorruptedDB(TypoDBError):
        pass

    def __str__(self):
        return "UserTypoDB ({})".format(self._user)

    def __init__(self, user, debug_mode=False): # TODO CHANGE to False
        assert is_user(user), "User {!r} does not exists".format(user)

        # Disable Typtop for root
        # assert pwd.getpwnam(user).pw_uid != 0,\
        #     "Currently Typtop is disabled for {} user.".format(user)

        self._user = user  # this is a real user.
        # homedir = pwd.getpwnam(self._user).pw_dir
        typo_dir = os.path.join(SEC_DB_PATH, user)
        self._db_path = os.path.join(typo_dir, DB_NAME + '.json')
        self._log_path = os.path.join(LOG_DIR, DB_NAME + '.log')
        # First thing first -- setting the logger object
        setup_logger(self._log_path, debug_mode, user)
        self.logger = logger  # just in case I need it
        logger.info("---UserTypoDB instantiated---")

        # creating dir only if it doesn't exist
        if not os.path.exists(typo_dir):
            try:
                os.makedirs(typo_dir)
                os.system(
                    "chgrp {1} {0} && chmod -R g+w {0} && chmod -R o-rw {0} {0}".format(typo_dir, GROUP)
                )
            except OSError as error:
                logger.error("Trying to create: {}, but seems like the database"
                             " is not initialized.".format(typo_dir))
                raise UserTypoDB.NoneInitiatedDB(error)

        if not os.path.exists(self._db_path):
            with open(self._db_path, 'w') as dbf:
                json.dump({}, dbf)
            # cmd = 'chown root:{1} {0} && chmod o-rw {0};'.format(self._db_path, GROUP)
            change_db_ownership(self._db_path)

        try:
            self._db = json.load(open(self._db_path, 'r'))
        except (ValueError, IOError) as e:
            self._db = {}

        if auxT in self._db:
            self._aux_tab = self._db[auxT]
        else:
            self._aux_tab = self._db[auxT] = {}

        # always contains the serialized versin of sk, pk
        self._sk, self._pk = None, None
        # the global salt for the hmac-id only will be available if
        # correct pw is provided.
        self._hmac_salt, self._pw, self._pwent = None, None, None

    def __del__(self):
        tmp_f = self._db_path + '.tmp'
        with open(tmp_f, 'wb') as f:
            json.dump(self._db, f, indent=2)
            f.flush(); os.fsync(f.fileno())
        os.rename(tmp_f, self._db_path)
        change_db_ownership(self._db_path)
        logger.info("---UserTypoDB deleted---")

    def init_typtop(self, pw, allow_typo_login=True):
        """Create the 'typtop' database in user's home-directory.  Changes

        Also, it initializes the required tables as well as the reuired
        variables, such as, the typo-cache size, the global salt etc.

        """
        logger.info("Initiating typtop db with {}".format(
            dict(allow_typo_login=allow_typo_login)
        ))
        change_db_ownership(self._db_path)

        # db[auxT].delete()         # make sure there's no old unrelevent data
        # doesn't delete log because it will also be used
        # whenever a password is changed

        # *************** Initializing Aux Data *************************
        # *************** add org password, its' pks && global salt: ********
        # 1. derive public_key from the original password
        # 2. encrypt the global salt with the enc pk
        install_id = get_machine_id()
        install_time = time.strftime('%Y-%m-%d %H:%M:%S', time.localtime())
        last_sent_time = get_time()
        self._hmac_salt = os.urandom(SALT_LENGTH) # global salt
        self._pk, self._sk = generate_key_pair()  # ECC key pair
        self._sk = serialize_sk(self._sk)
        self._pw = pw
        self._aux_tab.update({
            INSTALLATION_ID: install_id,
            INSTALLATION_DATE: install_time,
            LOG_LAST_SENTTIME: last_sent_time,
            LOG_SENT_PERIOD: UPDATE_GAPS,
            SYSTEM_STATUS: SYSTEM_STATUS_NOT_INITIALIZED,
            LOGIN_COUNT: 0,
            ALLOWED_TYPO_LOGIN: allow_typo_login,
            ALLOWED_UPLOAD: True,
            ENC_PK: serialize_pk(self._pk),
            INDEX_J: random.randint(0, WAITLIST_SIZE-1),
        })

        # Just get the ids of all possible typo candidates for warming
        # up the cache.
        perm_index = self._fill_cache_w_garbage()
        if WARM_UP_CACHE:
            freq_counts = range(CACHE_SIZE, 0, -1)
            for i in range(CACHE_SIZE):
                freq_counts[perm_index[i]] = freq_counts[i]
        else:
            freq_counts = [0 for _ in range(CACHE_SIZE)]
        header_ctx = pkencrypt(self._pk, json.dumps({
            REAL_PW: self._pw,
            HMAC_SALT: urlsafe_b64encode(self._hmac_salt),
            FREQ_COUNTS: freq_counts
        }))

        logger.info("Initializing the auxiliary data base ({})".format(auxT))
        self._aux_tab[HEADER_CTX] = header_ctx

        # self._aux_tab.create_index(['desc'])
        self.set_status(SYSTEM_STATUS_ALL_GOOD)
        # 3. Filling the Typocache with garbage
        self._fill_waitlist_w_garbage()
        logger.debug("Initialization Complete")
        isON = self.get_from_auxtdb(ALLOWED_TYPO_LOGIN)   # isON: boolean
        logger.info("typtop is ON? {}".format(isON))

    def reinit_typtop(self, newPw):
        """
        Re-initiate the DB after a pw change.
        Most peripherial system settings don't change, including installID
        generates a new hmac salt,
        and encrypts the new pw, pw_ent, and the hmac salt
        """
        if not self.is_typtop_init():
            self.init_typtop(newPw)
        # Mostly a simple copy-paste of steps 1 to 2.5
        logger.info("Re-intializing after a pw change")
        # 1. derive public_key from the original password
        # 2. encrypt the global salt with the enc pk
        self._hmac_salt = os.urandom(16) # global salt
        pk, sk = generate_key_pair()  # ECC key pair
        self._pk, self._sk = pk, serialize_sk(sk)
        self._pw = newPw
        perm_index = self._fill_cache_w_garbage()
        if WARM_UP_CACHE:
            freq_counts = range(CACHE_SIZE, 0, -1)
            for i, f in enumerate(range(CACHE_SIZE)):
                freq_counts[perm_index[i]] = freq_counts[i]
        else:
            freq_counts = [0 for _ in range(CACHE_SIZE)]

        header_ctx = pkencrypt(self._pk, json.dumps({
            REAL_PW: self._pw,
            HMAC_SALT: urlsafe_b64encode(self._hmac_salt),
            FREQ_COUNTS: freq_counts
        }))
        self.set_in_auxtdb(HEADER_CTX, header_ctx)
        self.set_in_auxtdb(ENC_PK, serialize_pk(self._pk))
        # 3 sending logs and deleting tables:
        logger.debug("Sending logs, deleting tables")

        # 4. Filling the Typocache with garbage
        self._fill_waitlist_w_garbage()
        self.set_status(SYSTEM_STATUS_ALL_GOOD)
        logger.info("RE-Initialization Complete")

    def is_typtop_init(self):
        """
        Returns whether the typtop has been set (might be installed
        but not active)
        """
        if not os.path.exists(self._db_path):
            return False
        if self.get_from_auxtdb(HEADER_CTX):
            return True
        else:
            return False

    def getdb(self, tabname):
        if tabname in self._db:
            return self._db[tabname]
        else:
            self._db[tabname] = []
            return self._db[tabname]

    def get_db_path(self):
        return self._db_path

    def assert_initialized(self):
        if not self.is_typtop_init():
            raise UserTypoDB.NoneInitiatedDB(
                "Typtop DB wasn't initiated yet!"
            )

    def is_allowed(self, what):
        self.assert_initialized()
        isON = self.get_from_auxtdb(what)
        assert isON in (True, False), \
            'Corrupted data in {}: {}={} ({})'.format(
                auxT, what, isON, type(isON)
            )
        return isON

    def allow(self, what, how):
        self.assert_initialized()
        assert how in (True, False, 0, 1), "Expects a boolean"
        how = True if how else False
        self.set_in_auxtdb(what, how)
        state = "ON" if how else "OFF"
        logger.info("typtop::{} set to {}".format(what, state))

    def allow_upload(self, allow):
        self.allow(ALLOWED_UPLOAD, allow)

    def allow_login(self, allow=True):
        self.allow(ALLOWED_TYPO_LOGIN, allow)

    def is_allowed_login(self):
        return self.is_allowed(ALLOWED_TYPO_LOGIN)

    def is_allowed_upload(self):
        return self.is_allowed(ALLOWED_UPLOAD)

    def is_real_pw(self, pw):
        return self._pw == pw

    def _fill_waitlist_w_garbage(self):
        ts = get_time()
        install_id = self.get_installation_id()

        def randomstring(k):
            return urlsafe_b64encode(os.urandom(16))

        waitlist = [
            pkencrypt(
                self._pk,
                json.dumps([install_id+randomstring(5), ts])
            )
            for _ in range(WAITLIST_SIZE)
        ]
        self.set_in_auxtdb(WAIT_LIST, waitlist)

    def _fill_cache_w_garbage(self):
        logger.debug("Filling Typocache with garbage")
        perm_index = range(CACHE_SIZE)
        random.shuffle(perm_index)
        pw = self._pw
        popular_typos = [os.urandom(16) for _ in range(CACHE_SIZE)]
        self._pwent = entropy(self._pw)

        # if WARM_UP_CACHE: # No need to check, assumes always WARM_UP
        i = 0
        for tpw in warm_up_with(pw):
            if WARM_UP_CACHE and i < CACHE_SIZE and pw != tpw and tpw not in popular_typos:
                self.insert_log(typo=tpw, in_cache=True, ts=-1)
                popular_typos[perm_index[i]] = tpw
                i += 1
            elif pw != tpw:
                self.insert_log(typo=tpw, in_cache=False, ts=-1)

        popular_typos = [pw] + popular_typos
        garbage_list = [
            pwencrypt(tpw, self._sk) for tpw in popular_typos
        ]
        self.set_in_auxtdb(TYPO_CACHE, garbage_list)
        return perm_index

    def get_installation_id(self):
        self.assert_initialized()
        return self.get_from_auxtdb(INSTALLATION_ID)

    def get_last_unsent_logs_iter(self, force=False):
        """
        Check what was the last time the log has been sent,
        And returns whether the log should be sent
        """
        logger.debug("Getting last unsent logs")
        if not self.is_typtop_init():
            logger.debug("Could not send. Typtop not initiated")
            return False, iter([])
        if not self.is_allowed_upload():
            logger.info("Not sending logs because send status set to False")
            return False, iter([])
        last_sending = self.get_from_auxtdb(LOG_LAST_SENTTIME)  # , float)
        update_gap = self.get_from_auxtdb(LOG_SENT_PERIOD)  # , float)
        time_now = time.time()
        passed_enough_time = ((time_now - last_sending) >= update_gap)
        if not force and not passed_enough_time:
            logger.debug("Not enough time has passed ({}) to send new logs."
                         .format(str(last_sending)))
            return False, iter([])
        log_t = self._db[logT]
        try:
            new_logs = iter(log_t)  # .find(log_t.table.columns.ts >= last_sending)
            logger.info("Prepared new logs to be sent, from {} to {}".format(
                str(last_sending), str(time_now))
            )
            return True, new_logs
        except AttributeError:
            return False, iter([])

    def update_last_log_sent_time(self, sent_time=0, delete_old_logs=True):
        logger.debug("updating log sent time")
        if not sent_time:
            sent_time = get_time()
            logger.debug("generating new timestamp={} ".format(sent_time))
        self._db[auxT][LOG_LAST_SENTTIME] = float(sent_time)
        if delete_old_logs:
            logger.debug("deleting old logs")
            del self._db[logT][:]
            # while self._db[logT]:
            #     self._db[logT].pop()
            # try:
            #     log_t.table.delete().where(
            #         log_t.table.columns.ts <= float(sent_time)
            #     ).execute()
            # except AttributeError:
            #     pass

    def insert_log(self, typo, in_cache, ts=None):
        """Updates the log with information about typo. Remember, if sk_dict is
        not provided it will insert @typo as typo_id and 0 as relative_entropy.
        Note the default values used in other_info, which is basically what
        is expected for the original password.
        """
        assert self._pw and self._hmac_salt
        # Only log columns:
        # log_columns = set([
        #     'tid', 'edit_dist', 'rel_entropy', 'ts',
        #     'istop5fixable', 'in_cache'
        # ])
        log_info = {
            'tid': compute_id(self._hmac_salt, typo),
            'edit_dist': distance(str(self._pw), str(typo)),
            'rel_entropy': entropy(typo) - self._pwent,
            'ts': ts if ts else get_time(),
            'istop5fixable': is_in_top5_fixes(self._pw, typo),
            'in_cache': in_cache
        }
        try:
            self._db[logT].append(log_info)
        except KeyError:
            self._db[logT] = [log_info]

    def _add_typo_to_waitlist(self, typo):
        """Adds the typo to the waitlist.
        @typo (string) : typo of the user's passwrod
        """
        logger.debug("Adding a new typo to waitlist")
        logger.debug("Adding: {}".format(typo))
        waitlist = self.get_from_auxtdb(WAIT_LIST)  # , yaml.load)
        indexj = int(self.get_from_auxtdb(INDEX_J))  # , int))
        ts = get_time()
        assert indexj < len(waitlist), \
            "Index_j={}, wait-list={}".format(indexj, waitlist)
        waitlist[indexj] = pkencrypt(self.get_pk(), json.dumps([typo, ts]))
        indexj = (indexj + 1) % WAITLIST_SIZE
        self.set_in_auxtdb(WAIT_LIST, waitlist)
        self.set_in_auxtdb(INDEX_J, indexj)
        logger.debug("Typo encrypted.")

    def _decrypt_n_filter_waitlist(self):
        """Decrypts the waitlist and filters out the ones failed validity
        check. After that it combines the typos and returns a list of
        typos sorted by their frequency.

        return: [(typo_i, f_i)]
        """
        filtered_typos = defaultdict(int)
        sk = deserialize_sk(self._sk)
        assert self._pwent, "PW is not initialized: {}".format(self._pwent)
        ignore = set()
        install_id = self.get_installation_id()
        for typo_ctx in self.get_from_auxtdb(WAIT_LIST): # , yaml.load):
            typo_txt = pkdecrypt(sk, typo_ctx)
            typo, ts = yaml.safe_load(typo_txt)
            # starts with installation id, then must be garbage
            if typo.startswith(install_id):
                continue
            self.insert_log(typo, in_cache=False, ts=ts)
            if typo in ignore: continue
            if self.validate(self._pw, typo):
                filtered_typos[typo] += 1
            else:
                logger.debug("Ignoring: {}".format(typo))
                ignore.add(typo)
        logger.info("Waitlist decrypted successfully")
        return sorted(
            filtered_typos.items(), key=lambda a: a[1], reverse=True
        )

    def get_table_size(self, table_name):
        return self._db[table_name].count()

    @staticmethod
    def get_typo_cache_size():
        return CACHE_SIZE

    def get_pk(self):
        """Returns the public key"""
        if not self._pk:
            self._pk = deserialize_pk(self.get_from_auxtdb(ENC_PK))
        return self._pk

    # TODO: More policies
    @staticmethod
    def cache_insert_policy(old_t_c, new_t_c):
        if old_t_c < 0:  # for garbage rows in cache
            return True
        d = old_t_c + new_t_c
        assert d > 0
        rnd = random.randint(0, d-1)
        return rnd <= new_t_c

    def clear_waitlist(self):
        self._fill_waitlist_w_garbage()
        logger.info("Waitlist is deleted.")

    def check_login_count(self, update=False):
        """Keeps track of how many times the user has successfully logged in."""
        count_entry = self.get_from_auxtdb(LOGIN_COUNT) + 1
        if update:
            self.set_in_auxtdb(LOGIN_COUNT, count_entry)
        allowed = count_entry > NUMBER_OF_ENTRIES_TO_ALLOW_TYPO_LOGIN
        if not allowed:
            logger.info("Checking login count. Allowed = {}".format(allowed))
        return allowed

    def check(self, pw):
        logger.info("Checking entered password.")
        pk = self.get_pk()   # cannot be tampered
        typo_cache = self.get_from_auxtdb(TYPO_CACHE)  # , yaml.load)
        match_found = False
        freq_counts = []
        i = 0
        for i, sk_ctx in enumerate(typo_cache):
            try:
                sk = pwdecrypt(pw, sk_ctx)
                if not verify_pk_sk(pk, bytes(sk)):  # Somehow the hash matched !!
                    logger.error("pk-sk Verification failed!!")
                    continue
                self._sk = sk
            except (TypeError, ValueError) as e: # Decryption Failed
                # print("Failed for {}.{} ({})".format(pw, i, e))
                # that means password did not match.
                continue
            header = yaml.safe_load(
                pkdecrypt(sk, self.get_from_auxtdb(HEADER_CTX))
            )
            self._hmac_salt = urlsafe_b64decode(header[HMAC_SALT])
            freq_counts = header[FREQ_COUNTS]
            if i > 0:
                freq_counts[i-1] += 1
            self._pw = header[REAL_PW]
            self._pwent = entropy(self._pw)
            self.insert_log(pw, in_cache=True, ts=get_time())
            match_found = True
            break
        allowed = False
        if match_found:
            assert self._pwent, "PW is not initialized: {}".format(self._pwent)
            self._update_typo_cache_by_waitlist(typo_cache, freq_counts)
            if i == 0:   # the real password entered
                self.check_login_count(update=True)
                allowed = True
            else:   # A typo entered
                allowed = (self.check_login_count(update=False) and
                           self.is_allowed_login())
        else:
            self._add_typo_to_waitlist(pw)
            allowed = False
        logger.info("TypTop's decision: LoginAllowed = {}".format(allowed))
        return allowed

    def get_from_auxtdb(self, key):
        return self._aux_tab.get(key, '')

    def set_in_auxtdb(self, key, value):
        self._aux_tab[key] = value

    def validate(self, orig_pw, typo):
        # Edit distance 1 is always allowed for all passwords; more
        # allowance if len(orig_pw) is large
        edit_dist = (distance(str(orig_pw), str(typo))-1)/float(len(orig_pw))
        typo_ent = entropy(typo)
        not_much_weaker = (typo_ent >= (self._pwent - REL_ENT_CUTOFF))
        not_too_weak = (typo_ent >= LOWER_ENT_CUTOFF)
        close_edit = (edit_dist <= EDIT_DIST_CUTOFF)
        return not_too_weak and not_much_weaker and close_edit

    def _update_typo_cache_by_waitlist(self, typo_cache, freq_counts):
        """
        Updates the hash cache according to waitlist and clears the waitlist
        @typo_cache: a list of typos in the cache, and @freq_counts are the
        corresponding frequencies.

        returns: Updated typo_cache and their frequencies. Also applies the
        permutations.
        """
        logger.info("Updating TypoCache by Waitlist")
        good_typo_list = self._decrypt_n_filter_waitlist()
        mini, minf = min(enumerate(freq_counts), key=itemgetter(1))
        for typo, f in good_typo_list:
            if UserTypoDB.cache_insert_policy(minf, f):
                logger.debug("Inserting: {} @ {}".format(typo, mini))
                typo_cache[mini+1] = pwencrypt(typo, self._sk)
                freq_counts[mini] = max(minf + 1, f)  # TODO: Check
                mini, minf = min(enumerate(freq_counts), key=itemgetter(1))
            else:
                logger.debug("I miss you: {} ({} <-> {})".format(typo, minf, f))
        # write the new typo_cache and freq_list
        # TODO: Apply permutation
        header_ctx = pkencrypt(self._pk, json.dumps({
            REAL_PW: self._pw,
            HMAC_SALT: urlsafe_b64encode(self._hmac_salt),
            FREQ_COUNTS: freq_counts
        }))
        logger.debug("Real pw={!r}".format(self._pw))

        self.set_in_auxtdb(HEADER_CTX, header_ctx)
        self.set_in_auxtdb(TYPO_CACHE, typo_cache)
        self.clear_waitlist()

    def get_prompt(self):
        """This used to supply password prompt, but in new versions,
        this function is useless"""
        # Password prompts
        return {
            SYSTEM_STATUS_ALL_GOOD: 'aDAPTIVE pASSWORD',
            SYSTEM_STATUS_NOT_INITIALIZED: 'Please Initialize',
            SYSTEM_STATUS_PW_CHANGED: 'Please Re-initialize',
            SYSTEM_STATUS_CORRUPTED_DB: 'Corrupted DB!'
        }.get(self.get_from_auxtdb(SYSTEM_STATUS), '(Error!) Password')

    def set_status(self, status):
        self.set_in_auxtdb(key=SYSTEM_STATUS, value=status)


def check_system_status(typo_db):
    sys_stat_val = typo_db.get_from_auxtdb(SYSTEM_STATUS)
    # if reached here - db should be initiated updating the entry count
    if not sys_stat_val:  # if not found in table
        raise UserTypoDB.CorruptedDB(
            "ERROR: (check_system_status) Typtop DB is Corrupted."
        )
    if sys_stat_val == SYSTEM_STATUS_PW_CHANGED:  # password_changed
        raise ValueError(SYSTEM_STATUS_PW_CHANGED)
    if sys_stat_val == SYSTEM_STATUS_CORRUPTED_DB:  # corrupted_db
        raise ValueError(SYSTEM_STATUS_CORRUPTED_DB)


def on_correct_password(typo_db, password):
    # log the entry of the original pwd
    logger.info("-^-")
    try:
        if not typo_db.is_typtop_init():
            logger.error("Typtop DB wasn't initiated yet!")
            typo_db.init_typtop(password)
            # the initialization is now part of the installation process
        check_system_status(typo_db)
        # correct password but db fails to see it
        is_match = typo_db.check(password)
        if not is_match:
            logger.debug("Changing system status to {}.".format(
                SYSTEM_STATUS_PW_CHANGED
            ))
            typo_db.set_status(SYSTEM_STATUS_PW_CHANGED)
            logger.info("Changing the system status because: match={}".format(is_match))
    except (ValueError, KeyError) as e:
        # most probably - an error of decryption as a result of pw change
        typo_db.set_status(SYSTEM_STATUS_PW_CHANGED)
        logger.exception("Key error raised. Probably a failure in decryption. Re-initializing...")
        typo_db.reinit_typtop(password)
    except Exception as e:
        logger.exception(
            "Unexpected error while on_correct_password:\n{}\n"
            .format(e)
        )
    # In order to avoid locking out - always return true for correct password
    return True


def on_wrong_password(typo_db, password):
    logger.info("-\/-")
    is_match = False
    try:
        if not typo_db.is_typtop_init():
            logger.error("Typtop DB wasn't initiated yet!")
            # typo_db.init_typtop(password)
            return False
        check_system_status(typo_db)
        is_match = typo_db.check(password)
        # password has changed and this is an old password
        if is_match and typo_db.is_real_pw(password):
            logger.info("Password changed, old password entered. Re-initializing..")
            typo_db.reinit_typtop(urlsafe_b64encode(os.urandom(16)))
            return False
    except (ValueError, KeyError) as e:
        # probably  failure in decryption
        logger.exception("Error!! Probably failure in decryption. Re-initializing...")
        typo_db.reinit_typtop(password)
    except Exception as e:
        logger.exception("Unexpected error while on_wrong_password:\n{}\n"\
                         .format(e))
        print("TypToP is not initialized.\n $ sudo typtop --init")
    return is_match


def call_check(exchk_ret_val, user, password):
    ret = -1
    # usage = '<1 or 0> <username> <password>'
    if not is_user(user):
        ret = 1
    else:
        typo_db = UserTypoDB(user)
        exchk_ret_val = str(exchk_ret_val)
        if exchk_ret_val == '0':
            ret = int(not on_correct_password(typo_db, password))
        else:
            ret = int(not on_wrong_password(typo_db, password))
    return ret

if __name__ == "__main__":
    sys.stdout.write("{}".format(call_check(*sys.argv[1:])))




from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives import (
    serialization, hashes, hmac
)
from cryptography.hazmat.primitives.ciphers import (
    Cipher, algorithms, modes
)
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
from cryptography.hazmat.primitives import padding

import os, random
from base64 import urlsafe_b64encode, urlsafe_b64decode
import struct

HASH_ALGOS = {
    'sha1': hashes.SHA1(),
    'sha224': hashes.SHA224(),
    'sha256': hashes.SHA256(),
    'sha384': hashes.SHA384(),
    'sha512': hashes.SHA512()
}
HASH_CNT = 1000 # Number of hashes to compute one SHA256 takes 15 microsec,
SALT_LENGTH = 16 # Length for the Password salt for PBKDF
HASH_ALGO = 'sha256' # For PBKDF HMAC
IV_LENGTH = 12 # Length of GCM IV
TAG_LENGTH = 16 # Length of the GCM tag, truncate if larger than this


def hash256(*args):
    """short function for Hashing the arguments with SHA-256"""
    assert len(args)>0, "Should give at least 1 message"
    assert all(isinstance(m, (bytes, basestring)) for m in args), \
        "All inputs should be byte string"
    h = hashes.Hash(hashes.SHA256(), backend=default_backend())
    h.update(bytes(len(args)) + bytes(args[0]) + bytes(len(args[0])))
    for m in args[1:]:
        h.update(bytes(m))
        h.update(bytes(len(m)))
    h.update(bytes(len(args)))
    return h.finalize()


def hmac256(secret, m):
    h = hmac.HMAC(bytes(secret), hashes.SHA256(), backend=default_backend())
    h.update(bytes(m))
    return h.finalize()


def generate_key_pair():
    private_key = ec.generate_private_key(
        ec.SECP256R1(), default_backend()
    )
    public_key = private_key.public_key()
    return public_key, private_key


def _slow_hash(pw, sa):
    kdf = PBKDF2HMAC(
        algorithm=hashes.SHA256(),
        length=16,
        salt=sa,
        iterations=HASH_CNT,
        backend=default_backend()
    )
    key = kdf.derive(pw)
    return key


def verify_pk_sk(pk, sk):
    try:
        if isinstance(sk, (bytes, basestring)):
            sk = deserialize_sk(sk)
        if not isinstance(pk, (bytes, basestring)):
            pk = serialize_pk(pk)
        pkprime = serialize_pk(sk.public_key())
        return pkprime == pk
    except Exception as e:
        print("ERRROR: {}".format(e))
        print("sk: {}\npk:{}".format(sk, pk))
        raise(e)


def verify(pw, sa, h):
    """Verifies if h(pw, sa) == h
    returns k in case it matches, otherwise None.
    """
    sa = urlsafe_b64decode(bytes(sa))
    k = _slow_hash(pw, sa)
    hprime = hash256(k)
    if h == urlsafe_b64encode(hprime):
        return k
    else:
        return None


def harden_pw(pw):
    sa = os.urandom(SALT_LENGTH)
    k = _slow_hash(pw, sa)
    h = hash256(k)
    return (urlsafe_b64encode(sa), k, urlsafe_b64encode(h))


def serialize_pk(pk):
    if isinstance(pk, (basestring, bytes)):
        return pk
    elif isinstance(pk, ec.EllipticCurvePrivateKey):
        pk = pk.public_key()
    return pk.public_bytes(
        serialization.Encoding.PEM,
        format=serialization.PublicFormat.SubjectPublicKeyInfo
    )


def serialize_sk(sk):
    if isinstance(sk, (basestring, bytes)):
        return sk
    else:
        return sk.private_bytes(
            serialization.Encoding.PEM,
            format=serialization.PrivateFormat.PKCS8,
            encryption_algorithm=serialization.NoEncryption()
        )


def deserialize_sk(sk_s):
    if isinstance(sk_s, (bytes, basestring, unicode)):
        return serialization.load_pem_private_key(
            bytes(sk_s), password=None, backend=default_backend()
        )
    else:
        return sk_s


def deserialize_pk(pk_s):
    if isinstance(pk_s, (bytes, basestring, unicode)):
        return serialization.load_pem_public_key(
            bytes(pk_s), backend=default_backend()
        )
    else:
        return pk_s


def pad_pw(pw, pad_length):
    """Pad pw to a pad_length, so that it hides the length of the password in bytes."""
    assert 0 < pad_length < 256
    pw = bytes(pw)
    k = len(pw)/pad_length
    topad = pw[k*pad_length:]
    topad_len = pad_length - len(topad)
    if topad_len == 0:
        topad_len = pad_length
    pad = chr(topad_len) * topad_len
    return pw + pad
    # padder = padding.PKCS7(pad_length*8).padder()
    # return padder.update(bytes(pw)) + padder.finalize()


def unpad_pw(padded_pw, pad_length):
    """Unpad pw"""
    padded_pw = bytes(padded_pw)
    padlen = ord(padded_pw[-1])
    assert padlen>0, "Malformed padding. Last byte cannot be zero."
    pad = padded_pw[-padlen:]
    assert all((padi==chr(padlen) for padi in pad))
    return padded_pw[:-padlen]

    # unpadder = padding.PKCS7(pad_length*8).unpadder()
    # unpadder.update(bytes(pw)) + unpadder.finalize()


def pwencrypt(pw, m):
    """Encrypt the message m under pw using AES-GCM method (AEAD scheme).
    iv = 0   # Promise me you will never reuse the key
    c = <hash_style>.<iteration>.<urlsafe-base64 <salt><iv><tag><ctx>>
    :hash_style: sha-256 or sha-512, scrypt
    :iteration: Number of iteration. These two are the parameters
    for PBKDF2HMAC.
    Size of the ciphertext:
    """
    m = m.encode('ascii', errors='ignore')
    hash_func = HASH_ALGOS[HASH_ALGO]
    itercnt = random.randint(HASH_CNT, 2*HASH_CNT)
    header_txt = HASH_ALGO + '.' + str(itercnt)
    sa = os.urandom(SALT_LENGTH)
    kdf = PBKDF2HMAC(
        algorithm=hash_func,
        length=16,
        salt=sa,
        iterations=itercnt,
        backend=default_backend()
    )
    key = kdf.derive(pw)

    iv, ctx, tag = _encrypt(key, m, associated_data=header_txt)
    # Salt (SALT_LENGTH), IV (IV_LENGTH), TAG (TAG_LENGTH)
    ctx_b64 = urlsafe_b64encode(sa + iv + tag + ctx)
    return header_txt + '.' + ctx_b64


def pwdecrypt(pw, full_ctx_b64):
    """
    Decrypt a ciphertext using pw,
    Recover, hash algo, iteration count, and salt, iv, tag, ctx from ctx_b64
    """
    full_ctx_b64 = full_ctx_b64.encode('ascii', errors='ignore')
    hash_algo, itercnt, ctx_b64 = full_ctx_b64.split('.')
    header_txt = hash_algo + '.' + itercnt
    ctx_bin = urlsafe_b64decode(ctx_b64)
    sa, ctx_bin = ctx_bin[:SALT_LENGTH], ctx_bin[SALT_LENGTH:]
    iv, ctx_bin = ctx_bin[:IV_LENGTH], ctx_bin[IV_LENGTH:]
    tag, ctx = ctx_bin[:TAG_LENGTH], ctx_bin[TAG_LENGTH:]
    kdf = PBKDF2HMAC(
        algorithm=HASH_ALGOS[hash_algo],
        length=16,
        salt=sa,
        iterations=int(itercnt),
        backend=default_backend()
    )
    key = kdf.derive(pw)
    try:
        m = _decrypt(key, iv, ctx, tag, associated_data=header_txt)
        return m
    except Exception as e:
        raise ValueError(e)


def _encrypt(key, plaintext, associated_data=''):
    # Generate a random 96-bit IV.
    iv = os.urandom(IV_LENGTH)
    if len(key) not in algorithms.AES.key_sizes:
        key = hash256(key)  # makes it 256-bit
    # Construct an AES-GCM Cipher object with the given key and a
    # randomly generated IV.
    encryptor = Cipher(
        algorithms.AES(key),
        modes.GCM(iv),
        backend=default_backend()
    ).encryptor()

    # associated_data will be authenticated but not encrypted,
    # it must also be passed in on decryption.
    encryptor.authenticate_additional_data(associated_data)

    # Encrypt the plaintext and get the associated ciphertext.
    # GCM does not require padding.
    ciphertext = encryptor.update(plaintext) + encryptor.finalize()

    return (iv, ciphertext, encryptor.tag)


def _decrypt(key, iv, ciphertext, tag, associated_data=''):
    # Construct a Cipher object, with the key, iv, and additionally the
    # GCM tag used for authenticating the message.
    if len(key) not in algorithms.AES.key_sizes:
        key = hash256(key)  # makes it 256-bit

    decryptor = Cipher(
        algorithms.AES(key),
        modes.GCM(iv, tag),
        backend=default_backend()
    ).decryptor()

    # We put associated_data back in or the tag will fail to verify
    # when we finalize the decryptor.
    decryptor.authenticate_additional_data(associated_data)

    # Decryption gets us the authenticated plaintext.
    # If the tag does not match an InvalidTag exception will be raised.
    return decryptor.update(ciphertext) + decryptor.finalize()


def encrypt(k, m):
    """Symmetric encrpyt.
    """
    if not isinstance(m, bytes):
        m = bytes(m)
    iv, ctx, tag = _encrypt(k, m)
    return urlsafe_b64encode(iv + ctx + tag)


def decrypt(k, ctx):
    """Symmetric decrpyt.
    """
    if not isinstance(ctx, bytes):
        ctx = bytes(ctx)
    ctx_bin = urlsafe_b64decode(bytes(ctx))
    iv, c, tag = ctx_bin[:12], ctx_bin[12:-16], ctx_bin[-16:]
    try:
        return _decrypt(k, iv, c, tag)
    except Exception as e:
        print(e)
        raise(ValueError(e))


def pkencrypt(pk, m):
    """Public key encrypt"""
    if not isinstance(pk, ec.EllipticCurvePublicKey):
        pk = deserialize_pk(pk)
    rpk, rsk = generate_key_pair()
    common_key = rsk.exchange(ec.ECDH(), pk)
    c = encrypt(common_key, m)
    rpk_s = serialize_pk(rpk)
    return '||'.join((rpk_s, c))


def pkdecrypt(sk, ctx):
    """Public key decrypt"""
    if isinstance(sk, (bytes, basestring)):
        sk = deserialize_sk(sk)
    rpk_s, c = ctx.split('||')
    rpk = deserialize_pk(rpk_s)
    common_key = sk.exchange(ec.ECDH(), rpk)
    m = decrypt(common_key, c)
    return m


def compute_id(salt, pwtypo):
    """
    Computes an ID for pwtypo.
    @pwtypo (byte string): mistyped (or correct) password
    @salt (byte string): global salt
    """
    if len(salt)>SALT_LENGTH:
        salt = urlsafe_b64decode(bytes(salt))
    assert len(salt) == SALT_LENGTH, "len(salt)={} ({})"\
        .format(len(salt), SALT_LENGTH)
    h = hmac256(salt, bytes(pwtypo))
    return struct.unpack('<I', h[:4])[0]


def encrpyt_sk(key, sk):
    if not isinstance(sk, (bytes, basestring)):
        sk = serialize_sk(sk)
    return encrypt(key, sk)




try: # try cryptography
    import cryptography
    # print "Uisng cryptogrpahy.io"
    from typtop.cryptography_pwpkcrypto import *
except ImportError as e:
    # print "Uisng pycryptodome", e
    from typtop.pycryptodome_pwpkcrypto import *



import sys
VERSION = "0.2.5"

DB_NAME = "typtop"
SEC_DB_PATH = '/etc/typtop.d'
LOG_DIR = '/var/log/'
# SEC_DB_NAME = DB_NAME + ".ro" # READ_ONLY // ROOT_ONL
BINDIR = '/usr/local/bin'  # careful pip install does not guarantee
                           # scripts to be installed at this location.
SYSTEM = ''


def set_distro():
    os = sys.platform
    if os == 'darwin':
        return 'darwin'
    elif os.startswith('linux'):
        try:
            import distro
            dist = distro.id()
            name = distro.name()
        except ImportError:
            dist = 'ubuntu' # default value, will remove DISTRO
            name = 'Ubuntu' # in future.
        # To add new distributions, refer to:
        #    http://distro.readthedocs.io/en/latest/#distro.id
        #    http://linuxmafia.com/faq/Admin/release-files.html
        if dist in ('ubuntu', 'debian'):
            return 'debian'
        elif dist in ('fedora', 'rhel', 'centos'):
            return 'fedora'
        elif dist == 'arch':
            return 'arch'
        else:
            raise ValueError(
                "Not supported for your Linux distribution: {}"\
                .format(name)
            )
    else:
        raise ValueError(
            "Not supported for your OS: {}"\
            .format(os)
        )

DISTRO = set_distro()


def warm_up_with(pw):
    return [
        pw.swapcase(), pw[0].swapcase()+pw[1:],
        pw + '1','1' + pw,
        '`' + pw, pw + '`',
        pw + '0', '0' + pw,
        pw[:-1] + pw[-1] + pw[-1]
    ]

# The group
GROUP = 'shadow' if DISTRO in ('debian') else \
    'root' if DISTRO in ('fedora', 'arch') else \
    'wheel' if DISTRO in ('darwin') \
    else ''

if sys.platform=='darwin':
    SYSTEM = 'OSX'
elif sys.platform.startswith('linux'):
    SYSTEM = 'LINUX'
else:
    raise ValueError("Not yet suporrted. Report in @github/rchatterjee/pam_typopw")

if SYSTEM == 'OSX':
    SEC_DB_PATH = '/usr/local/etc/typtop.d/' # ETC is not writable due to SIP in OSX
elif SYSTEM == 'LINUX':
    SEC_DB_PATH = '/usr/local/etc/typtop.d/'  # Changing from /etc/pam_typtop


################################################################################
                  ########## PARAMETERS ##########
################################################################################
# default values
CACHE_SIZE = 5  # Size of the typo cache
WAITLIST_SIZE = 10   # Size of the waitlist
PADDED_PW_SIZE = 64  # length of the padded passwords
EDIT_DIST_CUTOFF = 1.0/10  # fractional distance between the real
                           # password and the typo, 1 typo per 10
                           # characters
REL_ENT_CUTOFF = 3 # Typo cannot be less than 3 bits in stregth
                   # compared to the real password
LOWER_ENT_CUTOFF = 0  # No point having something like this
NUMBER_OF_ENTRIES_TO_ALLOW_TYPO_LOGIN = 0  # Number of successful login   (# TODO: set to 30 when done testing)
NUMBER_OF_DAYS_TO_ALLOW_TYPO_LOGIN = 15  # Number of days since installation
UPDATE_GAPS = 6 * 60 * 60 # 6 hours, in seconds
WARM_UP_CACHE = 1  # Should the cache be warmed up or not.
################################################################################


# column names in sqlite tables
auxT = 'Header' # holds most of the system's setting
# in ENCRYPTED format
HEADER_CTX = 'HeaderCtx'
HMAC_SALT = 'HMACSalt'
FREQ_COUNTS = 'FreqCounts'
REAL_PW = 'RealPassword'

# in PLAINTEXT format
ENC_PK = 'EncPublicKey'
INDEX_J = 'IndexJ'
# for book keeping
ALLOWED_TYPO_LOGIN = 'AllowedTypoLogin'
LOGIN_COUNT = 'LoginCount'
INSTALLATION_ID = 'InstallationId'
INSTALLATION_DATE = 'InstallationDate'
ALLOWED_LOGGING = 'AllowLogging'
ALLOWED_UPLOAD = 'AllowUpload'
LOG_LAST_SENTTIME = 'LastLogSetntTime'
LOG_SENT_PERIOD = 'PeriodForSendingLog' # How frequently to send the logs
SYSTEM_STATUS = 'SystemStatus'
SYSTEM_STATUS_PW_CHANGED = 'PasswordChanged'
SYSTEM_STATUS_ALL_GOOD = 'StatusAllGood'
SYSTEM_STATUS_CORRUPTED_DB = 'StatusCorruptedDB'
SYSTEM_STATUS_NOT_INITIALIZED = 'NotInitialized'

# Typo Cache and Wait list are stored as just another data field in
# Header. No need to have different tables for them.
TYPO_CACHE = "TypoCache"  # PW Encryption of sk
WAIT_LIST = 'WaitList'  # PK encryption of typo and timestamp

# ORIG_PW = 'OrignalPw'
# ORIG_PW_ID = 'OrgPwID'  # Remove
# ORIG_PW_ENTROPY_CTX = 'OrgignalPwEntropyCtx'
# ORIG_PW_SGN_PK = 'SgnPublicKey'
# ORIG_SGN_SALT = 'OriginalPwSaltForVerifySecretKey'
# REL_ENT_CUTOFF = "RelativeEntropyDecreaseAllowed"
# MIN_ENT_CUTOFF = "LowestEntropyAllowed"
# COUNT_KEY_CTX = "CountKeyCtx"
# HMAC_SALT_CTX = 'HMACSaltCtx'
#
#
#
logT = 'Log'
logT_cols = [
    'tid', 'edit_dist', 'rel_entropy', 'ts',
    'istop5fixable', 'in_cache'
]

GITHUB_URL = 'https://github.com/rchatterjee/pam-typopw'  # URL in github repo

first_msg = """\n\n
  /  |                          /  |
 _$$ |_    __    __   ______   _$$ |_     ______    ______
/ $$   |  /  |  /  | /      \ / $$   |   /      \  /      \\
$$$$$$/   $$ |  $$ |/$$$$$$  |$$$$$$/   /$$$$$$  |/$$$$$$  |
  $$ | __ $$ |  $$ |$$ |  $$ |  $$ | __ $$ |  $$ |$$ |  $$ |
  $$ |/  |$$ \__$$ |$$ |__$$ |  $$ |/  |$$ \__$$ |$$ |__$$ |
  $$  $$/ $$    $$ |$$    $$/   $$  $$/ $$    $$/ $$    $$/
   $$$$/   $$$$$$$ |$$$$$$$/     $$$$/   $$$$$$/  $$$$$$$/
          /  \__$$ |$$ |                          $$ |
          $$    $$/ $$ |                          $$ |
           $$$$$$/  $$/                           $$/
Hello!

Thanks for installing TypToP (version: {version}).  This software
attaches a new pluggable authentication module (PAM) to some of your
common authentication processes, such as su, login, screensaver etc.,
and observes for password typing mistakes. It records your frequent
typing mistakes, and enable logging in with slight vairations of your
actual login password that are frequent and safe to do so.

This is a research prototype, and we are collecting some anonymous
non-sensitive data about your password typing patterns to verify our
design. The details of what we collect, how we collect and store, and
the security blueprint of this software can be found in the GitHub
page: {url}.  The participation in the study is completely voluntary,
and you can opt out at any time while still keep using the software.

Checkout other options (such as opting out of the study) of the
utility script typtop by running:

$ typtops.py --help

Note, You have to initiate this for each user who intend to use the
benefit of adaptive typo-tolerant password login.
""".format(url=GITHUB_URL, version=VERSION)



#!/usr/bin/env python

import os
import sys
import random
import requests
import json
from typtop.dbaccess import UserTypoDB, get_time
from typtop.config import LOG_DIR, VERSION, SEC_DB_PATH, DB_NAME

# Disable this weird warning. (Future TODO - deal with this.)
from requests.packages.urllib3.exceptions import SubjectAltNameWarning
requests.packages.urllib3.disable_warnings(SubjectAltNameWarning)

# note - there's no way this script will be called
# without the DB being initialized, because we call it
# AFTER a SUCCESSFUL login

THIS_FOLDER = os.path.dirname(os.path.abspath(__file__))
CERT_FILE = os.path.join(THIS_FOLDER, 'typtopserver.crt')


def send_logs(typo_db, force=False):
    need_to_send, iter_data = typo_db.get_last_unsent_logs_iter(force)
    logger = typo_db.logger
    if not need_to_send:
        logger.info("No need to send logs now.")
        return

    list_of_logs = list(iter_data)
    install_id = str(typo_db.get_installation_id())
    dbdata = json.dumps(list_of_logs)
    url = 'https://ec2-54-209-30-18.compute-1.amazonaws.com/submit'
    r = requests.post(
        url,
        data=dict(
            # urlsafe-base64 does not have '#'
            uid=install_id.strip() + '#' + str(VERSION),
            data=dbdata,
            test=0,
        ),
        allow_redirects=True,
        verify=CERT_FILE
    )
    sent_successfully = (r.status_code == 200)
    logger.info("Sent logs status {} ({}) (sent_successfully={})"
                .format(r.status_code, r.text, sent_successfully))
    # deletes the logs that we have sent
    if sent_successfully:
        typo_db.update_last_log_sent_time(
            sent_time=get_time(),
            delete_old_logs=True
        )
        # truncate log file to last 200 lines and look for update if available
        updatecmd = "typtop --update" if random.randint(0, 100) <= 20 else ''
        cmd = """
        tail -n500 {0}/{1}.log > /tmp/t.log && mv /tmp/t.log {0}/{1}.log;
        {2}
        """.format(LOG_DIR, DB_NAME, updatecmd)
        os.system(cmd)


def main():
    assert len(sys.argv) > 1
    user = sys.argv[1]
    users = [user]
    force = True if (len(sys.argv) > 2 and sys.argv[2] == 'force') else False
    if user == 'all':  # run for all users
        users = [
            d for d in os.listdir(SEC_DB_PATH)
            if os.path.isdir(os.path.join(SEC_DB_PATH, d))
        ]
    for user in users:
        typo_db = UserTypoDB(user)
        send_logs(typo_db, force)

if __name__ == '__main__':
    main()



from __future__ import print_function

# import os
# from subprocess import Popen, call, PIPE
# from glob import glob
# try:
from setuptools import setup

# from setuptools.command.install import install
# except ImportError as e:
#     print(e)
#     print("Setuptools not found, if installation fails, "
#           "please install setuptools, and try again.")
#     from distutils.core import setup
#     from distutils.command.install import install

from typtop.config import (
    VERSION, set_distro, first_msg, GITHUB_URL
)


SCRIPTS = {
    'typtop/send_typo_log.py',
    'typtop/typtops.py'
}

DISTRO = set_distro()

LIB_DEPENDENCIES = {
    'debian': [
        'libffi-dev', 'python-pkg-resources', 'libssl-dev',
        'python-setuptools', 'python-dev',
    ],
    # Fedora does not have python-pam!! So, we cannot write the fedora
    # version.
    'fedora': [
        'libffi-devel', 'openssl-devel',
        'python-devel', 'python-pip', 'python-setuptools',
    ],
    'darwin': [],
    'arch'  : [],
}[DISTRO]

PACMAN = {
    'debian': 'apt-get install -y'.split(),
    'fedora': 'yum install -y'.split(),
    'darwin': [],
    'arch': [],
}[DISTRO]

PYTHON_DEPS = [
    'cryptography',
    # 'pycryptodome',
    'word2keypress',
    # 'dataset',
    'zxcvbn',
    'requests==2.11.1', # 2.12 has issue with x509 cetificate
    'pyyaml',
    'distro',
]

OPTIONS = {
    'argv_emulation': True,
    # 'packages': ['requests', 'requests', 'selenium']
}

# Shitty python way of getting these files
# DATA_FILES = [
#     ('/tmp/typtop_osx/pam_opendirectory', [
#         "osx/pam_opendirectory/pam_opendirectory_typo.c",
#         "osx/pam_opendirectory/typtops.c",
#         "osx/pam_opendirectory/run_as_root.c",
#         "osx/pam_opendirectory/Makefile"
#     ]),
#     ('/tmp/typtop_linux/unixchkpwd', [])
# ]
# With the help from http://peterdowns.com/posts/first-time-with-pypi.html
setup(
    name='typtop',   # 'loginwitherror',
    # app=['typtop/dbaccess.py'],
    packages=['typtop'], # this must be the same as the name above
    version=VERSION,
    description='Adaptive typo-tolerant password checking for Debian logins',
    long_description=first_msg,
    author='Rahul Chatterjee, Yuval Pnueli',
    author_email='rc737@cornell.edu',
    url=GITHUB_URL,
    download_url='{}/tarball/{}'.format(GITHUB_URL, VERSION),
    keywords=[
        'Password', 'typo-tolerance',
        'login-with-errors', 'Login'
    ],
    # scripts=SCRIPTS,
    entry_points={
        'console_scripts': [
            'typtops.py = typtop.typtops:main',
            'send_typo_log.py = typtop.send_typo_log:main']
    },
    license='MIT',
    package_dir={'typtop': 'typtop/'},
    package_data={
        'typtop': ['../LICENSE', '../README.md', 'typtopserver.crt'],
    },
    data_files=[],  # DATA_FILES,
    include_package_data=True,
    options={'py2app': OPTIONS},
    classifiers=['Development Status :: 4 - Beta'],
    install_requires=PYTHON_DEPS,
    # cmdclass={'install': CustomInstaller},
    zip_safe=True,
)




import struct
from typtop.pw_pkcrypto import decrypt, hash256, urlsafe_b64encode
import logging
from typtop.config import DB_NAME, GROUP
import pwd, grp
import uuid
import os


def is_user(u):
    try:
        pwd.getpwnam(u)
        return True
    except KeyError:
        return False

logger = logging.getLogger(DB_NAME)


def change_db_ownership(fl):
    try:
        g_id = grp.getgrnam(GROUP).gr_gid
        f_stat = os.stat(fl)
        f_uid, f_gid, mode = f_stat.st_uid, f_stat.st_gid, f_stat.st_mode
        # 511 = 2**9 - 1, last 3x3 permission bits, Should be 0660
        # o-rw
        if f_uid != 0 or f_gid != g_id or (mode & 511 != 0o660):
            os.chown(fl, f_uid, g_id)
            os.chmod(fl, 0o660)
    except (KeyError, OSError) as e:
        logger.exception(e)


def setup_logger(logfile_path, debug_mode, user):
    log_level = logging.DEBUG if debug_mode else logging.INFO
    logger.setLevel(log_level)
    if not logger.handlers:  # if it doesn't have an handler yet:
        handler = logging.FileHandler(logfile_path)
        formatter = logging.Formatter(
            '%(asctime)s:%(levelname)s:<{}>:[%(filename)s:%(lineno)s'\
            '(%(funcName)s)>> %(message)s'.format(user)
        )
        handler.setFormatter(formatter)
        logger.addHandler(handler)


def is_in_top5_fixes(orig_pw, typo):
    return orig_pw in (
        typo.capitalize(), typo.swapcase(), typo.lower(),
        typo.upper(), typo[1:], typo[:-1]
    )


def increment_val(dbh, tabname, key, keyf='desc', valuef='data'):
    """
    Increments the value of the key in tabname purely using sql query.
    final value of key in tabname will be tabname[keyf=key] + 1
    """
    q = "update {tabname} set {valuef}=(select {valuef}+1 from {tabnme} "\
        "where desc=\"{key}\") where {keyf}=\"{key}\"".format(
            tabname=tabname, keyf=keyf, key=key, valuef=valuef
        )
    dbh.query(q)


def get_machine_id():
    """Unique machine id tied to the machine, does not change easily"""
    return urlsafe_b64encode(hash256(bytes(uuid.getnode()))[:6])


def find_one(table, key, apply_type=str):
    """Finds a key from the table with column name 'desc', and value in
    'data'. It is faster than traditional find_one of 'dataset'.

    """
    # q = 'select data from {} where desc="{}" limit 1'.format(
    #     table.table.name, key
    # )
    # res = ''
    return apply_type(table.get(key, ''))


def decode_decrypt_sym_count(key, ctx):
    """
    Receives the count ctx, decrypts it, decode it from base64
    and than from bytes to int
    """
    count_in_bytes = decrypt(key, ctx)
    # raise error if bigger? TODO
    return struct.unpack('<i',count_in_bytes)[0]
